/* ModelPseudoAdiabaticMovingSectional.hpp 
 * 
 * Copyright (C) 2010 Sylwester Arabas
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or (at
 * your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef MODEL_PSEUDOADIABATIC_MOVING_SECTIONAL_HPP
#  define MODEL_PSEUDOADIABATIC_MOVING_SECTIONAL_HPP

#  include "drops.hpp"

/* does not work... :(
#  include "/usr/local/src/sundials-2.4.0/src/cvode/cvode_direct_impl.h"
*/

#  define self ModelPseudoAdiabaticMovingSectional
class self : public Model
{

  private: class params : public ModelParams
  {
    public: Updraft *U;
    public: InBinSpectrum *I;
    public: SpectraMemLayout *ML;
    public: SatVapPresMltplr *SVPM;
    public: DiffCoeffMltplr *DCM;

    public: params(Updraft* U_, InBinSpectrum* I_, SpectraMemLayout* ML_, 
      SatVapPresMltplr* SVPM_, DiffCoeffMltplr* DCM_, vector<Solute*> *S_) 
    {
      U = U_; 
      I = I_;
      ML = ML_;
      SVPM = SVPM_;
      DCM = DCM_;
      S = S_;
    }
  };

  public: ModelParams* getParams(Updraft* U, InBinSpectrum* I, 
    SpectraMemLayout* ML, SatVapPresMltplr* SVPM, DiffCoeffMltplr* DCM, vector<Solute*>* S) 
  {
    return new self::params(U, I, ML, SVPM, DCM, S);
  } 

#  define idx_p 0
#  define idx_T 1
#  define idx_q_v 2
#  define idx_r 3
 
  public: size_t getStateVectorIndexPressure() { return idx_p; }
  public: size_t getStateVectorIndexTemperature() { return idx_T; }
  public: size_t getStateVectorIndexSpecificHumidity() { return idx_q_v; }
  public: size_t getStateVectorIndexRadii() { return idx_r; }
  public: bool isBulk() { return false; }

#  define P (static_cast<self::params*>(P_))
#  define t (t_ * si::seconds)
#  define q_v NV_Ith_S(Y, idx_q_v) 
#  define p (NV_Ith_S(Y, idx_p) * quantity<si::pressure>(1 * si::pascals))
#  define T (NV_Ith_S(Y, idx_T) * quantity<si::temperature>(1 * si::kelvins))
#  define r_d_l(s, i) (NV_Ith_S(P->pp, P->ML->r_d_l_ix(s, i)) * si::metres)
#  define r_d_r(s, i) (NV_Ith_S(P->pp, P->ML->r_d_r_ix(s, i)) * si::metres)
#  define r_l(s, b) (NV_Ith_S(Y, idx_r + P->ML->r_l_ix(s, b)) * si::metres)
#  define r_r(s, b) (NV_Ith_S(Y, idx_r + P->ML->r_r_ix(s, b)) * si::metres)
#  define T_o_l(s, i) (NV_Ith_S(Y, idx_r + P->ML->To_l_ix(s, i)) * si::kelvins)
#  define T_o_r(s, i) (NV_Ith_S(Y, idx_r + P->ML->To_r_ix(s, i)) * si::kelvins)
#  define N(s, b) (NV_Ith_S(P->pp, P->ML->N_ix(s, b)) / si::kilogrammes)

  // TODO: This is an approximate analytical formula for the Jacobian
  //       (and could be more procise!) 
  public: CVDlsDenseJacFn getODEJacFnPtr() { return &self::jacobian; }
  public: static int jacobian(int N, realtype t_, N_Vector Y, N_Vector dY_dt,	 
    DlsMat Jac, void *P_, N_Vector tmp1, N_Vector tmp2, N_Vector tmp3)
  {

/* intended for debugging purposes... but it does not work... :(
    int result = cvDlsDenseDQJac(N, t_, Y, dY_dt, Jac, P_, tmp1, tmp2, tmp3);
*/

    // Jac is initially zeroed by CVODE

#  define F_p (NV_Ith_S(dY_dt, idx_p) * si::pascals / si::seconds)
#  define F_q_v (NV_Ith_S(dY_dt, idx_q_v) / si::seconds)
#  define F_r_l(s, i) (NV_Ith_S(dY_dt, idx_r + P->ML->r_l_ix(s, i)) * si::metres / si::seconds)
#  define F_r_r(s, i) (NV_Ith_S(dY_dt, idx_r + P->ML->r_l_ix(s, i)) * si::metres / si::seconds)
#  define F_T_o_l(s, i) (NV_Ith_S(dY_dt, idx_r + P->ML->To_l_ix(s, i)) * si::kelvins / si::seconds)
#  define F_T_o_r(s, i) (NV_Ith_S(dY_dt, idx_r + P->ML->To_l_ix(s, i)) * si::kelvins / si::seconds)

    // \partial_T
    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; i < P->ML->n_bins(s) + 1; ++i)
      {
#  define J_r_T_unit (si::metres / si::kelvins / si::seconds)
#  define J_r_l_T(s, i) DENSE_ELEM(Jac, idx_r + P->ML->r_l_ix(s, i), idx_T)
        J_r_l_T(s, i) = (
          - p * P->D_0 / pow<2>(T) / P->rho_w_0 / r_l(s, i) / P->R_d / (
            1. / q_v + 1. / P->epsilon - 1
          )
        ) / J_r_T_unit;

#  define J_T_o_T_unit (1. / si::seconds)
#  define J_T_o_l_T(s, i) DENSE_ELEM(Jac, idx_r + P->ML->To_l_ix(s, i), idx_T)
        J_T_o_l_T(s,i) = (
          3. / (2. * P->c_p_w - P->c_p_v) * (
            (P->l_v(T_o_l(s,i)) - P->c_p_w * T_o_l(s,i)) / r_l(s,i) * J_r_l_T(s,i) * J_r_T_unit
            + P->K_0 / P->rho_w_0 * pow<-2>(r_l(s, i)) 
          )
        ) / J_T_o_T_unit;
#  undef J_T_o_l_T
#  undef J_T_o_T_unit
      }
    }

#  define J_r_r_T(s, i) DENSE_ELEM(Jac, idx_r + P->ML->r_r_ix(s, i), idx_T)
#  define J_T_o_r_T(s, i) DENSE_ELEM(Jac, idx_r + P->ML->To_r_ix(s, i), idx_T)

    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; P->ML->bin_r(s, i) != -1; i = P->ML->bin_r(s, i))
      {
#  define J_q_v_T DENSE_ELEM(Jac, idx_q_v, idx_T)
#  define J_q_v_T_unit (1. / si::kelvins / si::seconds)
        J_q_v_T += (
          (q_v - 1.) * M_PI * P->rho_w_0 / 3. * N(s, i) * (
            P->I->kappa(r_r(s, i), r_l(s, i)) * J_r_r_T(s, i) * J_r_T_unit + 
            P->I->kappa(r_l(s, i), r_r(s, i)) * J_r_l_T(s, i) * J_r_T_unit 
          )
        ) / J_q_v_T_unit;
#  undef J_r_l_T
#  undef J_r_r_T
#  undef J_r_T_unit
      }
    }

#  define J_p_T DENSE_ELEM(Jac, idx_p, idx_T)
#  define J_p_T_unit (si::pascals / si::kelvins / si::seconds)
    J_p_T = (
      - F_p / T
    ) / J_p_T_unit;

#  define J_T_T DENSE_ELEM(Jac, idx_T, idx_T)
#  define J_T_T_unit (1. / si::seconds)
    J_T_T = (
      P->R(q_v) / p / P->c_p(q_v) * (F_p + T * J_p_T * J_p_T_unit)
      - (P->l_v(T) * J_q_v_T * J_q_v_T_unit + F_q_v * (P->c_p_v - P->c_p_w)) / P->c_p(q_v)
    ) / J_T_T_unit;
#  undef J_T_T
#  undef J_T_T_unit

#  undef J_q_v_T
#  undef J_q_v_T_unit
#  undef J_p_T
#  undef J_p_T_unit

    // \partial_r_o
    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; i < P->ML->n_bins(s) + 1; ++i)
      {
#  define J_r_r_unit (1. / si::seconds)
#  define J_r_l_r_l(s, i, ss, ii) DENSE_ELEM(Jac, idx_r + P->ML->r_l_ix(s, i), idx_r + P->ML->r_l_ix(ss, ii))
        J_r_l_r_l(s, i, s, i) = (
          - (
            F_r_l(s, i) + P->D_0 / P->rho_w_0 / P->R_v / T_o_l(s, i) * P->p_v_s(T_o_l(s, i))
            * P->SVPM->d_mltplr__d_r(T_o_l(s, i), r_l(s, i), r_d_l(s, i), P->S->at(s))
          ) / r_l(s, i)
        ) / J_r_r_unit;

#  define J_T_o_r_unit (si::kelvins / si::metres / si::seconds)
#  define J_T_o_l_r_l(s, i, ss, ii) DENSE_ELEM(Jac, idx_r + P->ML->To_l_ix(s, i), idx_r + P->ML->r_l_ix(ss, ii))
        J_T_o_l_r_l(s, i, s, i) = (
          - F_T_o_l(s, i) / r_l(s, i)  
          + 3. / (2. * P->c_p_w - P->c_p_v) / r_l(s, i) * (
            J_r_l_r_l(s, i, s, i) * J_r_r_unit * (P->l_v(T_o_l(s, i)) - P->c_p_w * T_o_l(s, i))
            - P->K_0 / P->rho_w_0 * pow<-2>(r_l(s, i)) * (T - T_o_l(s, i))
          )
        ) / J_T_o_r_unit;
#  undef J_T_o_l_r_l
#  undef J_T_o_r_r_r
#  undef J_T_o_r_unit
      }
    }

#  define J_r_r_r_r(s, i, ss, ii) DENSE_ELEM(Jac, idx_r + P->ML->r_r_ix(s, i), idx_r + P->ML->r_r_ix(ss, ii))
#  define J_T_o_r_r_r(s, i, ss, ii) DENSE_ELEM(Jac, idx_r + P->ML->To_r_ix(s, i), idx_r + P->ML->r_r_ix(ss, ii))

    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; P->ML->bin_r(s, i) != -1; i = P->ML->bin_r(s, i))
      {
#  define J_q_v_r_unit (1. / si::metres / si::seconds)
#  define J_q_v_r_l(s, i) DENSE_ELEM(Jac, idx_q_v, idx_r + P->ML->r_l_ix(s, i))
        J_q_v_r_l(s, i) += (
          (q_v - 1.) * M_PI * P->rho_w_0 / 3. * N(s, i) * (
            F_r_r(s, i) * P->I->d_kappa__d_b(r_r(s, i), r_l(s, i)) +
            F_r_l(s, i) * P->I->d_kappa__d_a(r_l(s, i), r_r(s, i)) +
            P->I->kappa(r_l(s, i), r_r(s, i)) * J_r_l_r_l(s, i, s, i) * J_r_r_unit
          )         
        ) / J_q_v_r_unit;
#  undef J_r_l_T_o_l
#  undef J_r_l_T_o_r
#  undef J_r_r_T_o_l
#  undef J_r_r_T_o_r
#  undef J_r_T_o_unit
      }
    }

#  define J_q_v_r_r(s, i) DENSE_ELEM(Jac, idx_q_v, idx_r + P->ML->r_r_ix(s, i))

    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; i < P->ML->n_bins(s) + 1; ++i)
      {
#  define J_T_r_unit (si::kelvins / si::metres / si::seconds)
#  define J_T_r_l(s, i) DENSE_ELEM(Jac, idx_T, idx_r + P->ML->r_l_ix(s, i)) 
        J_T_r_l(s, i) = (
          P->l_v(T) / P->c_p(q_v) * J_q_v_r_l(s, i) * J_q_v_r_unit
        ) / J_T_r_unit;
#  undef J_T_r_l
#  undef J_T_r_r
#  undef J_T_r_unit
      }
    }

#  define J_T_r_r(s, i) DENSE_ELEM(Jac, idx_T, idx_r + P->ML->r_r_ix(s, i)) 

#  undef J_q_v_r_l
#  undef J_q_v_r_r
#  undef J_q_v_r_unit

    // \partial_T_o

    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; i < P->ML->n_bins(s) + 1; ++i)
      {
#  define J_r_T_o_unit (si::metres / si::kelvins / si::seconds)
#  define J_r_l_T_o_l(s, i, ss, ii) DENSE_ELEM(Jac, idx_r + P->ML->r_l_ix(s, i), idx_r + P->ML->To_l_ix(ss, ii))
        J_r_l_T_o_l(s, i, s, i) = (
          P->D_0 / P->rho_w_0 / r_l(s, i) / T_o_l(s, i) * (
            P->p_v_s(T_o_l(s, i)) * P->SVPM->p_v_s_r__p_v_s_infty(
              T_o_l(s, i), r_l(s, i), r_d_l(s, i), P->S->at(s)
            ) / P->R_v / T_o_l(s, i)
            - P->l_v(T_o_l(s, i)) * pow<-2>(P->R_v * T_o_l(s, i)) * p * q_v / (P->epsilon + q_v * (1. - P->epsilon))
          )
        ) / J_r_T_o_unit;

#  define J_T_o_T_o_unit (1. / si::seconds)
#  define J_T_o_l_T_o_l(s, i, ss, ii) DENSE_ELEM(Jac, idx_r + P->ML->To_l_ix(s, i), idx_r + P->ML->To_l_ix(ss, ii))
        J_T_o_l_T_o_l(s, i, s, i) = (
          3. / (2. * P->c_p_w - P->c_p_v) / r_l(s, i) * (
            J_r_l_T_o_l(s, i, s, i) * J_r_T_o_unit * (P->l_v(T_o_l(s, i)) - P->c_p_w * T_o_l(s, i))
            + F_r_l(s, i) * (P->c_p_v - 2. * P->c_p_w)
            - P->K_0 / P->rho_w_0 / r_l(s, i)
          )
        ) / J_T_o_T_o_unit;
#  undef J_T_o_l_T_o_l
#  undef J_T_o_T_o_unit        

      }
    }

#  define J_T_o_r_T_o_r(s, i, ss, ii) DENSE_ELEM(Jac, idx_r + P->ML->To_r_ix(s, i), idx_r + P->ML->To_r_ix(ss, ii))

    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; P->ML->bin_r(s, i) != -1; i = P->ML->bin_r(s, i))
      {
#  define J_q_v_T_o_unit (1. / si::kelvins / si::seconds)
#  define J_q_v_T_o_l(s, i) DENSE_ELEM(Jac, idx_q_v, idx_r + P->ML->To_l_ix(s, i))
        J_q_v_T_o_l(s, i) = (
          (q_v - 1.) * M_PI * P->rho_w_0 / 3. * N(s, i) 
          * P->I->kappa(r_l(s, i), r_r(s, i)) * J_r_l_T_o_l(s, i, s, i) * J_r_T_o_unit
        ) / J_q_v_T_o_unit;
      }
    }
#  undef J_r_l_T_o_l
#  undef J_r_r_T_o_r
#  undef J_T_o_r_unit

#  define J_q_v_T_o_r(s, i) DENSE_ELEM(Jac, idx_q_v, idx_r + P->ML->To_r_ix(s, i))

    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; i < P->ML->n_bins(s) + 1; ++i)
      {
#  define J_T_T_o_unit (1. / si::seconds)
#  define J_T_T_o_l(s, i) DENSE_ELEM(Jac, idx_T, idx_r + P->ML->To_l_ix(s, i)) 
        J_T_T_o_l(s, i) = (
          - P->l_v(T) / P->c_p(q_v) * J_q_v_T_o_l(s, i) * J_q_v_T_o_unit
        ) / J_T_T_o_unit;
#  undef J_T_T_o_l
#  undef J_T_T_o_unit
      }
    }
 

#  undef J_q_v_T_o_l
#  undef J_q_v_T_o_r
#  undef J_q_v_T_o_unit

    // \partial_q_v

    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; i < P->ML->n_bins(s) + 1; ++i)
      {
#  define J_r_q_v_unit (si::metres / si::seconds)
#  define J_r_l_q_v(s, i) DENSE_ELEM(Jac, idx_r + P->ML->r_l_ix(s, i), idx_q_v)
        J_r_l_q_v(s, i) = (
          P->D_0 / P->rho_w_0 / r_l(s, i) / P->R_d / T * p * pow<-2>(q_v) 
          / (1. / q_v + 1. / P->epsilon - 1.)
        ) / J_r_q_v_unit;

#  define J_T_o_q_v_unit (si::kelvins / si::seconds)
#  define J_T_o_l_q_v(s, i) DENSE_ELEM(Jac, idx_r + P->ML->To_l_ix(s, i), idx_q_v)
        J_T_o_l_q_v(s, i) = (
          3. / (2. * P->c_p_w - P->c_p_v) * J_r_l_q_v(s, i) * J_r_q_v_unit * (
            P->l_v(T_o_l(s, i)) - P->c_p_w * T_o_l(s, i)
          ) / r_l(s, i)
        ) / J_T_o_q_v_unit;
#  undef J_T_o_l_q_v 
#  undef J_T_o_q_v_unit        
      }
    }

#  define J_r_r_q_v(s, i) DENSE_ELEM(Jac, idx_r + P->ML->r_r_ix(s, i), idx_q_v)

#  define J_q_v_q_v_unit (1. / si::seconds)
#  define J_q_v_q_v DENSE_ELEM(Jac, idx_q_v, idx_q_v)
    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; P->ML->bin_r(s, i) != -1; i = P->ML->bin_r(s, i))
      { 
        J_q_v_q_v += (
          (q_v - 1.) * M_PI * P->rho_w_0 / 3. * N(s, i) * (
            P->I->kappa(r_r(s, i), r_l(s, i)) * J_r_r_q_v(s, i) * J_r_q_v_unit + 
            P->I->kappa(r_l(s, i), r_r(s, i)) * J_r_l_q_v(s, i) * J_r_q_v_unit 
          )
        ) / J_q_v_q_v_unit;
      }
    }
    J_q_v_q_v += (
      F_q_v / q_v  
    ) / J_q_v_q_v_unit;

#  undef J_r_l_q_v
#  undef J_r_r_q_v
#  undef J_r_q_v_unit

#  define J_p_q_v_unit (si::pascals / si::seconds)
#  define J_p_q_v DENSE_ELEM(Jac, idx_p, idx_q_v)
    J_p_q_v = (
      - F_p / P->R(q_v) * (P->R_v - P->R_d)
    ) / J_p_q_v_unit;

#  define J_T_q_v_unit (si::kelvins / si::seconds)
#  define J_T_q_v DENSE_ELEM(Jac, idx_T, idx_q_v)
    J_T_q_v = (
      T / p / P->c_p(q_v) * (
        F_p * (P->R_v - P->R_d)  
        - P->R(q_v) / P->c_p(q_v) * (P->c_p_v - P->c_p_d) * F_p 
        + P->R(q_v) * J_p_q_v * J_p_q_v_unit
      )
      - P->l_v(T) / P->c_p(q_v) * (
        J_q_v_q_v * J_q_v_q_v_unit 
        - F_q_v * (P->c_p_v - P->c_p_d) / P->c_p(q_v)
      )   
    ) / J_T_q_v_unit;
#  undef J_T_q_v
#  undef J_T_q_v_unit

#  undef J_q_v_q_v
#  undef J_q_v_q_v_unit
#  undef J_p_q_v
#  undef J_p_q_v_unit

    // \partial_p

    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; i < P->ML->n_bins(s) + 1; ++i)
      {
#  define J_r_p_unit (si::metres / si::pascals / si::seconds)
#  define J_r_l_p(s, i) DENSE_ELEM(Jac, idx_r + P->ML->r_l_ix(s, i), idx_p) 
        J_r_l_p(s, i) = (
          P->D_0 / P->rho_w_0 / r_l(s, i) / P->R_d / T / (1. / q_v + 1. / P->epsilon - 1.)
        ) / J_r_p_unit;

#  define J_T_o_p_unit (si::kelvins / si::pascals / si::seconds)
#  define J_T_o_l_p(s, i) DENSE_ELEM(Jac, idx_r + P->ML->To_l_ix(s, i), idx_p)
        J_T_o_l_p(s, i) = (
          3. / (2. * P->c_p_w - P->c_p_v) * (
            P->l_v(T_o_l(s, i)) - P->c_p_w * T_o_l(s, i)
          ) / r_l(s, i) * J_r_l_p(s, i) * J_r_p_unit
        ) / J_T_o_p_unit;
#  undef J_T_o_l_p
#  undef J_T_o_p_unit
      }
    }

#  define J_r_r_p(s, i) DENSE_ELEM(Jac, idx_r + P->ML->r_r_ix(s, i), idx_p)

    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; P->ML->bin_r(s, i) != -1; i = P->ML->bin_r(s, i))
      {
#  define J_q_v_p_unit (1. / si::pascals / si::seconds)
#  define J_q_v_p DENSE_ELEM(Jac, idx_q_v, idx_p)
        J_q_v_p += (
          (q_v - 1.) * M_PI * P->rho_w_0 / 3. * N(s, i) * (
            P->I->kappa(r_r(s, i), r_l(s, i)) * J_r_r_p(s, i) * J_r_p_unit +
            P->I->kappa(r_l(s, i), r_r(s, i)) * J_r_l_p(s, i) * J_r_p_unit
          )
        ) / J_q_v_p_unit;
      }
    }

#  undef J_r_l_p
#  undef J_r_r_p
#  undef J_r_p_unit

#  define J_p_p_unit (1. / si::seconds)
#  define J_p_p DENSE_ELEM(Jac, idx_p, idx_p)
    J_p_p = (
      F_p / p
    ) / J_p_p_unit;

#  define J_T_p_unit (si::kelvins / si::pascals / si::seconds)
#  define J_T_p DENSE_ELEM(Jac, idx_T, idx_p)
    J_T_p = (
      - P->l_v(T) / P->c_p(q_v) * J_q_v_p * J_q_v_p_unit
    ) / J_T_p_unit;
#  undef J_T_p
#  undef J_T_P_unit

#  undef J_p_p
#  undef J_p_p_unit
#  undef J_q_v_p
#  undef J_q_v_p_unit


#  undef F_p
#  undef F_q_v
#  undef F_r_l
#  undef F_r_r

    return 0;
  }

  // ODE system for SUNDIALS
  public: CVRhsFn getODERhsFnPtr() { return &self::system; }
  public: static int system(realtype t_, N_Vector Y, N_Vector dY_dt, void * P_)
  { 
    if (q_v < 0) 
    {
      cerr << msgprefix << "q_v < 0 ! -> asking for refinement..." << endl;
      return 1; 
    }

#  define P (static_cast<self::params*>(P_))

#  define dr_l__dt(s, i) NV_Ith_S(dY_dt, idx_r + P->ML->r_l_ix(s, i)) 
#  define dr_r__dt(s, i) NV_Ith_S(dY_dt, idx_r + P->ML->r_r_ix(s, i)) 
#  define dr__dt_unit (quantity<si::length>(1 * si::metres) / quantity<si::time>(1 * si::seconds))
    for (int s = 0; s < P->ML->n_specs(); ++s)
    {
      for (int i = 0; P->ML->bin_r(s, i) != -1; i = P->ML->bin_r(s, i))
      {
        if (r_l(s, i) > r_r(s, i)) 
        {
          cerr << "r_l (" << i << " -> " << r_l(s, i) << ") > r_r (" << r_r(s, i) << ")! -> asking for refinement..." << endl;
          return 1; 
        }

        if (r_d_l(s, i) > r_d_r(s, i)) 
        {
          cerr << "r_d_l (" << i << " -> " << r_d_l(s, i) << ") > r_d_r (" << r_d_r(s, i) << ")! -> asking for refinement..." << endl;
          return 1; 
        }
      }

      for (int i = 0; i < P->ML->n_bins(s) + 1; ++i)
      {
        if (r_l(s, i) < r_d_l(s, i)) 
        {
          cerr << msgprefix << "r_l (" << r_l(s, i) << ") < r_d_l (" << r_d_l(s, i) << ") ! -> asking for refinement..." << endl;
          return 1; 
        }

        {
          quantity<si::mass_density> rho_2r, rho_eq_l, rho_eq_r;

          rho_2r = q_v * p / P->R_d / T / (1. - q_v * (1. - 1. / P->epsilon));

          rho_eq_l = P->p_v_s(T_o_l(s, i)) * P->SVPM->p_v_s_r__p_v_s_infty(
            T_o_l(s, i), r_l(s, i), r_d_l(s, i), P->S->at(s)
          ) / P->R_v / T_o_l(s, i);

          dr_l__dt(s, i) = (
            (P->D_0 * P->DCM->D_transition__D_continuum(P->Kn_D(r_l(s, i), T_o_l(s, i)))) 
            / P->rho_w_0 * (rho_2r - rho_eq_l) / r_l(s, i)
          ) / dr__dt_unit;
        }

#  define dTo_l__dt(s, i) NV_Ith_S(dY_dt, idx_r + P->ML->To_l_ix(s, i))
#  define dTo_r__dt(s, i) NV_Ith_S(dY_dt, idx_r + P->ML->To_r_ix(s, i))
#  define dTo__dt_unit (si::kelvins / si::seconds)
        dTo_l__dt(s,i) = (
          3. / (P->c_p_w * r_l(s, i)) * (
            (P->K_0 * P->DCM->D_transition__D_continuum(P->Kn_K(r_l(s, i), T, p))) 
            * (T - T_o_l(s, i)) / r_l(s, i) / P->rho_w_0
            + dr_l__dt(s, i) * dr__dt_unit * P->l_v(T_o_l(s, i))
          )
        ) / dTo__dt_unit;
#  undef T_o_l
#  undef T_o_r
#  undef r
#  undef dr__dt
      }
    }

    {
      quantity<inverse_mass_density_over_time> sum(0 * si::cubic_metres / si::kilogrammes / si::seconds);
      for (int s = 0; s < P->ML->n_specs(); ++s)
      {
        for (int b = 0; P->ML->bin_r(s, b) != -1; b = P->ML->bin_r(s, b))
        {

          sum += N(s, b) * (
            P->I->kappa(r_r(s, b), r_l(s, b)) * dr_r__dt(s, b) * dr__dt_unit +
            P->I->kappa(r_l(s, b), r_r(s, b)) * dr_l__dt(s, b) * dr__dt_unit 
          );
#  undef dr__dt_unit
#  undef N
#  undef dr_l__dt
#  undef dr_r__dt
#  undef r_l
#  undef r_r
        }
      }
#  define dq_v__dt NV_Ith_S(dY_dt,idx_q_v) 
#  define dq_v__dt_unit (quantity<si::dimensionless>(1) / quantity<si::time>(1 * si::seconds))
      dq_v__dt = (
        (q_v - 1.) * M_PI * P->rho_w_0 / 3. * sum
      ) / dq_v__dt_unit;
   }

#  define dp__dt NV_Ith_S(dY_dt,idx_p) 
#  define dp__dt_unit (quantity<si::pressure>(1 * si::pascals) / quantity<si::time>(1 * si::seconds))
    dp__dt = ( 
      - p * P->g_0 * P->U->U(t) / P->R(q_v) / T
    ) / dp__dt_unit;

#  define dT__dt NV_Ith_S(dY_dt,idx_T)
#  define dT__dt_unit (quantity<si::temperature>(1 * si::kelvins) / quantity<si::time>(1 * si::seconds))
    dT__dt = (
      T * P->R(q_v) / p / P->c_p(q_v) * dp__dt * dp__dt_unit
      - P->l_v(T) / P->c_p(q_v) * dq_v__dt * dq_v__dt_unit
    ) / dT__dt_unit;

    return 0; // 0: ok, 1 :adjust-timestep, -1: fatal
  }

#  undef t
#  undef p
#  undef T
#  undef dp__dt
#  undef dp__dt_unit
#  undef dT__dt
#  undef dT__dt_unit
#  undef q_v
#  undef dq_v__dt
#  undef dq_v__dt_unit
#  undef P

#  undef idx_p
#  undef idx_T
#  undef idx_q_v

};

#  undef self
#endif
